﻿
using HeuristicLab.Algorithms.DataAnalysis.MctsSymbolicRegression;

namespace HeuristicLab.Algorithms.DataAnalysis.MctsSymbolicRegression {
  // only counts the number of variable references
  internal class SimpleConstraintHandler : IConstraintHandler {
    private int maxVariables;
    private int numVariables;

    public bool IsInvalidExpression {
      get {
        return numVariables > maxVariables;
      }
    }

    public SimpleConstraintHandler(int maxVars) {
      this.maxVariables = maxVars;
    }

    public void AddVarToCurrentFactor(int state) {
      numVariables++;
    }

    public void EndFactor() {
      // ignore
    }

    public void EndTerm() {
      // ignore
    }

    public bool IsAllowedFollowState(int currentState, int followState) {
      return numVariables < maxVariables ||
        // going to the final state is always allowed (smaller states are closer to the final state)
        currentState > followState;
        // (
        // followState != Automaton.StateExpFactorStart &&
        // followState != Automaton.StateFactorStart &&
        // followState != Automaton.StateTermStart &&
        // followState != Automaton.StateLogTStart &&
        // followState != Automaton.StateLogTFStart &&
        // followState != Automaton.StateInvTStart &&
        // followState != Automaton.StateInvTFStart);
    }

    public void Reset() {
      numVariables = 0;      
    }

    public void StartFactor(int state) {
      // ignore
    }

    public void StartNewTermInPoly() {
      // ignore
    }

    public void StartTerm() {
      // ignore
    }
  }
}
